use crate::Ledger;
use actix_web::{
    post,
    rt::time::{self, Interval},
    web,
    web::Bytes,
    HttpResponse, Responder,
};
use futures_util::Stream;
use std::{
    pin::Pin,
    task::{Context, Poll},
    time::Duration,
};
use tokio::sync::mpsc::Receiver;
use verax::Filter;

struct SubscriberStream {
    receiver: Receiver<verax::Transaction>,
    ping_interval: Interval,
    ping: u128,
}

impl Stream for SubscriberStream {
    type Item = Result<Bytes, actix_web::Error>;

    fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        match Pin::new(&mut self.ping_interval).poll_tick(cx) {
            Poll::Ready(_) => {
                // Send a heartbeat message
                self.ping += 1;
                let message = format!("{}\"ping\": {}{}\n", "{", self.ping, "}");
                let heartbeat_bytes = Bytes::copy_from_slice(&message.as_bytes());
                return Poll::Ready(Some(Ok(heartbeat_bytes)));
            }
            Poll::Pending => {}
        }

        match self.receiver.poll_recv(cx) {
            Poll::Ready(Some(msg)) => {
                let mut msg = serde_json::to_vec(&msg)?;
                msg.push(b'\n');
                Poll::Ready(Some(Ok(Bytes::copy_from_slice(&msg))))
            }
            Poll::Ready(None) => Poll::Ready(None),
            Poll::Pending => Poll::Pending,
        }
    }
}

#[post("/subscribe")]
pub async fn subscribe_by_tag(tag: web::Json<Filter>, ctx: web::Data<Ledger>) -> impl Responder {
    let (_, receiver) = ctx._inner.subscribe(tag.0).await;

    HttpResponse::Ok()
        .content_type("application/json")
        .streaming(SubscriberStream {
            receiver,
            ping_interval: time::interval(Duration::from_secs(30)),
            ping: 0,
        })
}
